/**
  This file belong to the KMPlayer project, a movie player plugin for Konqueror
  Copyright (C) 2007  Koos Vriezen <koos.vriezen@gmail.com>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**/

#include <config.h>

#include <map>
#include <cairo-xlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "kmplayertypes.h"
#include "kmplayer.h"
#include "viewarea.h"
#include "kmplayer_smil.h"
#include "kmplayer_rp.h"

unsigned char * pixbufBits (GdkPixbuf *pixbuf, int & channels);

using namespace KMPlayer;

//-------------------------------------------------------------------------

namespace KMPlayer {
    typedef std::map <String, ImageDataPtrW> ImageDataMap;
    static ImageDataMap * image_data_map;
}

ImageData::ImageData( const String & img) :
    cairo_image (0L),
    image (0L),
    url (img) {}

ImageData::~ImageData() {
    if (!url.isEmpty ())
        image_data_map->erase (url);
    if (cairo_image)
        cairo_pattern_destroy (cairo_image);
    delete image;
}

cairo_pattern_t *
ImageData::cairoImage (Single sw, Single sh, cairo_surface_t * similar) {
    if (cairo_image) {
        if (sw == w && sh == h)
            return cairo_image;
        cairo_pattern_destroy (cairo_image);
        cairo_image = 0L;
    }
    if (!image || sw <= 0 || sh <= 0)
        return 0L;
    int iw = image->width ();
    int ih = image->height ();
    w = sw;
    h = sh;
 // debugLog() << "cairoImage " << (int)width() << "x" << (int)height() << endl;
    const GPixbuf & gpb = (const GPixbuf) *image;
    int channels;
    unsigned char * bits = pixbufBits (gpb.pixbuf (), channels);
    cairo_surface_t * sf = cairo_image_surface_create_for_data (
            bits, channels == 4 ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24,
            iw, ih, iw*4);
    cairo_pattern_t * cp = cairo_pattern_create_for_surface (sf);
    cairo_surface_destroy (sf);
    cairo_pattern_set_extend (cp, CAIRO_EXTEND_NONE);
    cairo_matrix_t mat;
    cairo_matrix_init_scale (&mat, 1.0 * iw/w, 1.0 * ih/h);
    cairo_pattern_set_matrix (cp, &mat);

    cairo_surface_t * img_surface = cairo_surface_create_similar (similar,
            channels == 3 ? CAIRO_CONTENT_COLOR : CAIRO_CONTENT_COLOR_ALPHA,
            w, h);
    cairo_t * c = cairo_create (img_surface);
    cairo_set_source (c, cp);
    cairo_paint (c);
    cairo_pattern_destroy (cp);
    //gdk_cairo_set_source_pixbuf(c, gpb.pixbuf (), 0, 0);
    cairo_destroy (c);
    g_free (bits);
    cairo_image = cairo_pattern_create_for_surface (img_surface);
    cairo_surface_flush (img_surface);
    cairo_surface_destroy (img_surface);
    return cairo_image;
}

cairo_pattern_t * ImageData::cairoImage (cairo_surface_t * similar) {
    return cairoImage (width (), height (), similar);
}

bool ImageData::isEmpty () {
    return !(image || cairo_image);
}

Single ImageData::width () {
    if (image)
        return image->width ();
    return w;
}

Single ImageData::height () {
    if (image)
        return image->height ();
    return h;
}

void CachedImage::setUrl (const String & url) {
    if (url.isEmpty ()) {
        data = ImageDataPtr (new ImageData (url));
    } else {
        ImageDataMap::iterator i = image_data_map->find (url);
        if (i == image_data_map->end ()) {
            data = ImageDataPtr (new ImageData (url));
            //i = image_data_map->insert (url, ImageDataPtrW (data));
            (*image_data_map)[url] = ImageDataPtrW (data);
        } else {
            ImageDataPtr safe = i->second;
            data = safe;
        }
    }
}

//-------------------------------------------------------------------------

namespace KMPlayer {

class KMPLAYER_NO_EXPORT ViewSurface : public Surface {
public:
    ViewSurface (ViewArea * widget);
    ViewSurface (ViewArea * widget, NodePtr owner, const SRect & rect);
    ~ViewSurface ();

    void clear () { m_first_child = 0L; }

    SurfacePtr createSurface (NodePtr owner, const SRect & rect);
    void toScreen (Single & x, Single & y, Single & w, Single & h);
    void resize (const SRect & rect);
    void repaint (Single x, Single y, Single w, Single h);
    void video (Single x, Single y, Single w, Single h);

    ViewArea * view_widget;
};

} // namespace

KDE_NO_CDTOR_EXPORT ViewSurface::ViewSurface (ViewArea * widget)
  : Surface (SRect (0, 0, 342, 260)), //widget->width (), widget->height ())),
    view_widget (widget)
{}

KDE_NO_CDTOR_EXPORT
ViewSurface::ViewSurface (ViewArea * widget, NodePtr owner, const SRect & rect)
  : Surface (owner, rect), view_widget (widget) {}

KDE_NO_CDTOR_EXPORT ViewSurface::~ViewSurface() {
    //debugLog () << "~ViewSurface" << endl;
}

KDE_NO_EXPORT
SurfacePtr ViewSurface::createSurface (NodePtr owner, const SRect & rect) {
    SurfacePtr surface = new ViewSurface (view_widget, owner, rect);
    appendChild (surface);
    return surface;
}

KDE_NO_EXPORT void ViewSurface::resize (const SRect & ) {
    /*if (rect == nrect)
        ;//return;
    SRect pr = rect.unite (nrect); // for repaint
    rect = nrect;*/
}

KDE_NO_EXPORT
void ViewSurface::toScreen (Single & x, Single & y, Single & w, Single & h) {
    Matrix matrix (xoffset, yoffset, xscale, yscale);
    matrix.translate (bounds.x (), bounds.y ());
    for (SurfacePtr s = parentNode(); s; s = s->parentNode()) {
        matrix.transform(Matrix (s->xoffset, s->yoffset, s->xscale, s->yscale));
        matrix.translate (s->bounds.x (), s->bounds.y ());
    }
    matrix.getXYWH (x, y, w, h);
}

KDE_NO_EXPORT
void ViewSurface::repaint (Single x, Single y, Single w, Single h) {
    toScreen (x, y, w, h);
    view_widget->scheduleRepaint (x, y, w, h);
    //kdDebug() << "Surface::repaint x:" << (int)x << " y:" << (int)y << " w:" << (int)w << " h:" << (int)h << endl;
}

KDE_NO_EXPORT void ViewSurface::video (Single x, Single y, Single w, Single h) {
    toScreen (x, y, w, h);
    debugLog() << "Surface::video:" << background_color << " " << (background_color & 0xff000000) << endl;
    view_widget->setAudioVideoGeometry (x, y, w, h,
            (background_color & 0xff000000 ? &background_color : 0));
}

//-------------------------------------------------------------------------

static cairo_surface_t * cairoCreateSurface (GtkWidget * wid, int w, int h) {
    GdkDisplay * gdisplay = gtk_widget_get_display (wid);
    Display * display = gdk_x11_display_get_xdisplay (gdisplay);
    Window id = gdk_x11_drawable_get_xid (GDK_DRAWABLE (wid->window));
    return cairo_xlib_surface_create (display, id,
            DefaultVisual (display, DefaultScreen (display)), w, h);
    /*return cairo_xlib_surface_create_with_xrender_format (
            qt_xdisplay (),
            id,
            DefaultScreenOfDisplay (qt_xdisplay ()),
            XRenderFindVisualFormat (qt_xdisplay (),
                DefaultVisual (qt_xdisplay (),
                    DefaultScreen (qt_xdisplay ()))),
            w, h);*/
}

#define CAIRO_SET_SOURCE_RGB(cr,c)           \
    cairo_set_source_rgb ((cr),               \
            1.0 * (((c) >> 16) & 0xff) / 255, \
            1.0 * (((c) >> 8) & 0xff) / 255,  \
            1.0 * (((c)) & 0xff) / 255)

class KMPLAYER_NO_EXPORT CairoPaintVisitor : public Visitor {
    SRect clip;
    cairo_surface_t * cairo_surface;
    cairo_surface_t * cairo_surface_widget;
    Matrix matrix;
    void traverseRegion (SMIL::RegionBase * reg);
public:
    cairo_t * cr;
    CairoPaintVisitor (cairo_surface_t * cs, const SRect & rect);
    ~CairoPaintVisitor ();
    using Visitor::visit;
    void visit (Node * n);
    void visit (SMIL::Layout *);
    void visit (SMIL::Region *);
    void visit (SMIL::ImageMediaType *);
    void visit (SMIL::TextMediaType *);
    void visit (SMIL::Brush *);
    void visit (SMIL::RefMediaType *) {}
    void visit (SMIL::AVMediaType *) {}
    void visit (RP::Imfl *);
    void visit (RP::Fill *);
    void visit (RP::Fadein *);
    void visit (RP::Fadeout *);
    void visit (RP::Crossfade *);
    void visit (RP::Wipe *);
    void visit (RP::ViewChange *);
};

KDE_NO_CDTOR_EXPORT
CairoPaintVisitor::CairoPaintVisitor (cairo_surface_t * cs, const SRect & rect)
 : clip (rect),
#if 0
 cairo_surface_widget (cs) {
    cairo_surface = cairo_image_surface_create
        (CAIRO_FORMAT_RGB24, rect.width(), rect.height());
    cr = cairo_create (cairo_surface);
    cairo_rectangle (cr, 0, 0, rect.width(), rect.height());
    cairo_set_source_rgb (cr, 0, 0, 0);
    cairo_fill (cr);
    cairo_translate (cr, -rect.x(), -rect.y());
    //cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
    cairo_set_tolerance (cr, 0.5 );
#else
 cairo_surface (cs) {
    cr = cairo_create (cs);
    cairo_rectangle (cr, rect.x(), rect.y(), rect.width(), rect.height());
    cairo_clip (cr);
    cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
    cairo_set_tolerance (cr, 0.5 );
    cairo_push_group (cr);
    cairo_set_source_rgb (cr, 0, 0, 0);
    cairo_rectangle (cr, rect.x(), rect.y(), rect.width(), rect.height());
    cairo_fill (cr);
#endif
}

KDE_NO_CDTOR_EXPORT CairoPaintVisitor::~CairoPaintVisitor () {
#if 0
    cairo_pattern_t * pat = cairo_pattern_create_for_surface (cairo_surface);
    cairo_surface_destroy (cairo_surface);
    cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
    cairo_matrix_t mat;
    cairo_matrix_init_translate (&mat, -clip.x(), -clip.y());
    cairo_pattern_set_matrix (pat, &mat);
    cairo_t * wcr = cairo_create (cairo_surface_widget);
    cairo_set_source (wcr, pat);
    cairo_rectangle (wcr, clip.x(), clip.y(), clip.width(), clip.height());
    cairo_fill (wcr);
    cairo_destroy (wcr);
    cairo_pattern_destroy (pat);
#else
    cairo_pattern_t * pat = cairo_pop_group (cr);
    cairo_set_source (cr, pat);
    cairo_rectangle (cr, clip.x(), clip.y(), clip.width(), clip.height());
    cairo_fill (cr);
    cairo_pattern_destroy (pat);
#endif
    cairo_destroy (cr);
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (Node * n) {
    warningLog() << "Paint called on " << n->nodeName() << endl;
}

KDE_NO_EXPORT void CairoPaintVisitor::traverseRegion (SMIL::RegionBase * reg) {
    // next visit listeners
    NodeRefListPtr nl = reg->listeners (mediatype_attached);
    if (nl) {
        for (NodeRefItemPtr c = nl->first(); c; c = c->nextSibling ())
            if (c->data)
                c->data->accept (this);
    }
    if (reg->surface && reg->surface->node && reg->surface->node.ptr () != reg)
        reg->surface->node->accept (this);

    // finally visit children, accounting for z-order FIXME optimize
    NodeRefList sorted;
    for (NodePtr n = reg->firstChild (); n; n = n->nextSibling ()) {
        if (n->id != SMIL::id_node_region)
            continue;
        SMIL::Region * r = static_cast <SMIL::Region *> (n.ptr ());
        NodeRefItemPtr rn = sorted.first ();
        for (; rn; rn = rn->nextSibling ())
            if (r->z_order < convertNode <SMIL::Region> (rn->data)->z_order) {
                sorted.insertBefore (new NodeRefItem (n), rn);
                break;
            }
        if (!rn)
            sorted.append (new NodeRefItem (n));
    }
    for (NodeRefItemPtr r = sorted.first (); r; r = r->nextSibling ())
        r->data->accept (this);
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::Layout * reg) {
    //debugLog() << "Visit " << reg->nodeName() << endl;
    SMIL::RegionBase *rb = convertNode <SMIL::RegionBase> (reg->rootLayout);
    if (reg->surface && rb) {
        //cairo_save (cr);
        Matrix m = matrix;

        SRect rect = reg->surface->bounds;
        Single xoff = reg->surface->xoffset;
        Single yoff = reg->surface->yoffset;
        Single w = rect.width() - 2 * xoff, h = rect.height() - 2 * yoff;
        matrix.getXYWH (xoff, yoff, w, h);

        SRect clip_save = clip;
        clip = clip.intersect (SRect (xoff, yoff, w, h));

        rb->surface = reg->surface;
        rb->surface->background_color = rb->background_color;

        if (reg->surface && (reg->surface->background_color & 0xff000000)) {
            CAIRO_SET_SOURCE_RGB (cr, reg->surface->background_color);
            cairo_rectangle (cr, clip.x (), clip.y(),
                    clip.width (), clip.height ());
            cairo_fill (cr);
        }
        //cairo_rectangle (cr, xoff, yoff, w, h);
        //cairo_clip (cr);

        matrix = Matrix (reg->surface->xoffset, reg->surface->yoffset,
                reg->surface->xscale, reg->surface->yscale);
        matrix.transform (m);
        traverseRegion (reg);
        //cairo_restore (cr);
        matrix = m;
        clip = clip_save;

        rb->surface = 0L;
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::Region * reg) {
    //debugLog() << "Visit " << reg->nodeName()  << endl;
    SRect rect = reg->surface->bounds;

    Matrix m = matrix;
    Single x = rect.x (), y = rect.y (), w = rect.width(), h = rect.height();
    matrix.getXYWH (x, y, w, h);
    if (!clip.intersect (SRect (x, y, w, h)).isValid ())
        return;
    matrix = Matrix (rect.x(), rect.y(), 1.0, 1.0);
    matrix.transform (m);
    SRect clip_save = clip;
    clip = clip.intersect (SRect (x, y, w, h));
    cairo_save (cr);
    if (reg->surface && (reg->surface->background_color & 0xff000000)) {
        CAIRO_SET_SOURCE_RGB (cr, reg->surface->background_color);
        cairo_rectangle (cr, clip.x (), clip.y(),
                clip.width (), clip.height ());
        cairo_fill (cr);
    }
    //cairo_rectangle (cr, x, y, w, h);
    //cairo_clip (cr);
    traverseRegion (reg);
    cairo_restore (cr);
    matrix = m;
    clip = clip_save;
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::ImageMediaType * img) {
    //debugLog() << "Visit " << img->nodeName()  << endl;
    ImageRuntime * ir = static_cast <ImageRuntime *> (img->runtime ());
    ImageData * id = ir->cached_img.data.ptr ();
    SRect rect = ir->intrinsicBounds ();
    if (id && !id->isEmpty () &&
            img->keepContent (img) &&
            rect.width() > 0 &&
            id->width() > 0 && id->height() > 0) {
        img->cached_rect = rect;
        Single x = rect.x ();
        Single y = rect.y ();
        Single w = rect.width();
        Single h = rect.height();
        matrix.getXYWH (x, y, w, h);
        SRect clip_rect = clip.intersect (SRect (x, y, w, h));
        if (!clip_rect.isValid ())
            return;
        cairo_pattern_t * pat = id->cairoImage (w, h, cairo_surface);
        if (pat) {
            cairo_matrix_t mat;
            cairo_matrix_init_identity (&mat);
            cairo_matrix_translate (&mat, -x, -y);
            cairo_pattern_set_matrix (pat, &mat);
            cairo_pattern_set_filter (pat, CAIRO_FILTER_FAST);
            cairo_set_source (cr, pat);
            cairo_rectangle (cr, clip_rect.x (), clip_rect.y (),
                    clip_rect.width (), clip_rect.height ());
            if (img->trans_steps > img->trans_step) { // only fadein for now
                cairo_save (cr);
                cairo_clip (cr);
                cairo_paint_with_alpha (
                        cr, 1.0 * img->trans_step/img->trans_steps);
                cairo_restore (cr);
            } else
                cairo_fill (cr);
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::TextMediaType * txt) {
    TextRuntime * td = static_cast <TextRuntime *> (txt->runtime ());
    //debugLog() << "Visit " << txt->nodeName() << " " << td->font_size << endl;
    SMIL::RegionBase * rb = convertNode <SMIL::RegionBase> (txt->region_node);
    if (rb && rb->surface && txt->keepContent (txt)) {
        SRect rect = rb->surface->bounds;
        Single x, y, w = rect.width(), h = rect.height();
        td->sizes.calcSizes (txt, rect.width(), rect.height(), x, y, w, h);
        matrix.getXYWH (x, y, w, h);
        /* QTextEdit * edit = new QTextEdit;
        edit->setReadOnly (true);
        edit->setHScrollBarMode (QScrollView::AlwaysOff);
        edit->setVScrollBarMode (QScrollView::AlwaysOff);
        edit->setFrameShape (QFrame::NoFrame);
        edit->setFrameShadow (QFrame::Plain);
        edit->setGeometry (0, 0, w, h);
        if (edit->length () == 0)
            edit->setText (text);
        if (w0 > 0)
            font.setPointSize (int (1.0 * w * font_size / w0));
        edit->setFont (font);
        QRect rect = p.clipRegion (QPainter::CoordPainter).boundingRect ();
        rect = rect.intersect (QRect (xoff, yoff, w, h));
        QPixmap pix = QPixmap::grabWidget (edit, rect.x () - (int) xoff,
                rect.y () - (int) yoff, rect.width (), rect.height ());*/

        float scale = 1.0 * w / rect.width (); // TODO: make an image
        cairo_set_font_size (cr, scale * td->font_size);
        cairo_font_extents_t txt_fnt;
        cairo_font_extents (cr, &txt_fnt);
        String str = td->text;
        struct Line {
            Line (const String & ln) : txt (ln), next(0) {}
            String txt;
            cairo_text_extents_t txt_ext;
            Single xoff;
            Line * next;
        } *lines = 0, *last_line = 0;
        Single y1 = y;
        Single max_width;
        Single min_xoff = w;
        while (!str.isEmpty () && y1 < y + h) {
            int len = str.find (Char ('\n'));
            bool skip_cr = false;
            if (len > 1 && str[len-1] == Char ('\r')) {
                --len;
                skip_cr = true;
            }
            String para = len > -1 ? str.left (len) : str;
            Line * line = new Line (para);
            if (!lines)
                lines = line;
            else
                last_line->next = line;
            last_line = line;
            int ppos = 0;
            while (true) {
                cairo_text_extents (cr, (const char*)line->txt, &line->txt_ext);
                float frag = line->txt_ext.width > 0.1
                    ? w / line->txt_ext.width : 1.1;
                if (frag < 1.0) {
                    int br_pos = int (line->txt.length () * frag); //educated guess
                    while (br_pos > 0) {
                        line->txt.truncate (br_pos);
                        br_pos = line->txt.findRev (Char (' '));
                        if (br_pos < 1)
                            break;
                        line->txt.truncate (br_pos);
                        cairo_text_extents (cr, (const char*)line->txt, &line->txt_ext);
                        if (line->txt_ext.width < (double)w)
                            break;
                    }
                }
                if (line->txt_ext.width > (double)max_width)
                    max_width = line->txt_ext.width;

                if (td->halign == TextRuntime::align_center)
                    line->xoff = (w - Single (line->txt_ext.width)) / 2;
                else if (td->halign == TextRuntime::align_right)
                    line->xoff = w - Single (line->txt_ext.width);
                if (line->xoff < min_xoff)
                    min_xoff = line->xoff;

                y1 += Single (txt_fnt.height);
                ppos += line->txt.length () + 1;
                if (ppos >= para.length () || y1 > y + h)
                    break;

                line->next = new Line (para.mid (ppos));
                line = line->next;
                last_line = line;
            }
            if (len < 0)
                break;
            str = str.mid (len + (skip_cr ? 2 : 1));
        }
        if (td->bg_opacity) { // TODO real alpha
            CAIRO_SET_SOURCE_RGB (cr, td->background_color);
            cairo_rectangle(cr, x + min_xoff, y,
                    (double)max_width + txt_fnt.max_x_advance / 2,
                    y1 - y /*txt_fnt.height + txt_fnt.descent*/);
            cairo_fill (cr);
        }
        CAIRO_SET_SOURCE_RGB (cr, td->font_color);
        while (lines) {
            Line * line = lines;
            line->xoff += Single (txt_fnt.max_x_advance / 4);
            cairo_move_to (cr, x+line->xoff, y + Single (txt_fnt.ascent));
            cairo_show_text (cr, (const char*)line->txt);
            y += Single (txt_fnt.height);
            lines = lines->next;
            delete line;
        }
        //cairo_stroke (cr);
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::Brush * brush) {
    //debugLog() << "Visit " << brush->nodeName() << endl;
    SMIL::RegionBase * rb = convertNode <SMIL::RegionBase> (brush->region_node);
    if (rb && rb->surface) {
        SRect rect = rb->surface->bounds;
        Single x, y, w = rect.width(), h = rect.height();
        matrix.getXYWH (x, y, w, h);
        unsigned int color = Color (brush->param ("color")).rgb ();
        CAIRO_SET_SOURCE_RGB (cr, color);
        cairo_rectangle (cr, x, y, w, h);
        cairo_fill (cr);
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Imfl * imfl) {
    if (imfl->surface) {
        cairo_save (cr);
        Matrix m = matrix;
        Single xoff = imfl->surface->xoffset, yoff = imfl->surface->yoffset;
        Single w = imfl->surface->bounds.width() - 2 * xoff;
        Single h = imfl->surface->bounds.height() - 2 * yoff;
        matrix.getXYWH (xoff, yoff, w, h);
        cairo_rectangle (cr, xoff, yoff, w, h);
        //cairo_clip (cr);
        cairo_translate (cr, xoff, yoff);
        cairo_scale (cr, w/imfl->width, h/imfl->height);
        if (imfl->needs_scene_img)
            cairo_push_group (cr);
        for (NodePtr n = imfl->firstChild (); n; n = n->nextSibling ())
            if (n->state >= Node::state_began &&
                    n->state < Node::state_deactivated) {
                RP::TimingsBase * tb = convertNode<RP::TimingsBase>(n);
                switch (n->id) {
                    case RP::id_node_viewchange:
                        if (!(int)tb->srcw)
                            tb->srcw = imfl->width;
                        if (!(int)tb->srch)
                            tb->srch = imfl->height;
                        // fall through
                    case RP::id_node_crossfade:
                    case RP::id_node_fadein:
                    case RP::id_node_fadeout:
                    case RP::id_node_fill:
                    case RP::id_node_wipe:
                        if (!(int)tb->w)
                            tb->w = imfl->width;
                        if (!(int)tb->h)
                            tb->h = imfl->height;
                        n->accept (this);
                        break;
                }
            }
        if (imfl->needs_scene_img) {
            cairo_pattern_t * pat = cairo_pop_group (cr);
            cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
            cairo_set_source (cr, pat);
            cairo_paint (cr);
            cairo_pattern_destroy (pat);
        }
        cairo_restore (cr);
        matrix = m;
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Fill * fi) {
    //kdDebug() << "Visit " << fi->nodeName() << endl;
    CAIRO_SET_SOURCE_RGB (cr, fi->color);
    if ((int)fi->w && (int)fi->h) {
        cairo_rectangle (cr, fi->x, fi->y, fi->w, fi->h);
        cairo_fill (cr);
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Fadein * fi) {
    //debugLog() << "Visit " << fi->nodeName() << endl;
    if (fi->target && fi->target->id == RP::id_node_image) {
        ImageData *id=convertNode<RP::Image>(fi->target)->cached_img.data.ptr();
        if (id && !id->isEmpty ()) {
            Single sx = fi->srcx, sy = fi->srcy, sw = fi->srcw, sh = fi->srch;
            if (!(int)sw)
                sw = id->width();
            if (!(int)sh)
                sh = id->height();
            if ((int)fi->w && (int)fi->h && (int)sw && (int)sh) {
                cairo_pattern_t * pat = id->cairoImage (cairo_surface);
                if (pat) {
                    cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
                    cairo_matrix_t matrix;
                    cairo_matrix_init_identity (&matrix);
                    float scalex = 1.0 * sw / fi->w;
                    float scaley = 1.0 * sh / fi->h;
                    cairo_matrix_scale (&matrix, scalex, scaley);
                    cairo_matrix_translate (&matrix,
                            1.0*sx/scalex - (double)fi->x,
                            1.0*sy/scaley - (double)fi->y);
                    cairo_pattern_set_matrix (pat, &matrix);
                    cairo_save (cr);
                    cairo_rectangle (cr, fi->x, fi->y, fi->w, fi->h);
                    cairo_set_source (cr, pat);
                    cairo_clip (cr);
                    cairo_paint_with_alpha (cr, 1.0 * fi->progress / 100);
                    cairo_restore (cr);
                }
            }
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Fadeout * fo) {
    //kdDebug() << "Visit " << fo->nodeName() << endl;
    if (fo->progress > 0) {
        CAIRO_SET_SOURCE_RGB (cr, fo->to_color);
        if ((int)fo->w && (int)fo->h) {
            cairo_save (cr);
            cairo_rectangle (cr, fo->x, fo->y, fo->w, fo->h);
            cairo_clip (cr);
            cairo_paint_with_alpha (cr, 1.0 * fo->progress / 100);
            cairo_restore (cr);
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Crossfade * cf) {
    //debugLog() << "Visit " << cf->nodeName() << endl;
    if (cf->target && cf->target->id == RP::id_node_image) {
        ImageData *id=convertNode<RP::Image>(cf->target)->cached_img.data.ptr();
        if (id && !id->isEmpty ()) {
            Single sx = cf->srcx, sy = cf->srcy, sw = cf->srcw, sh = cf->srch;
            if (!(int)sw)
                sw = id->width();
            if (!(int)sh)
                sh = id->height();
            if ((int)cf->w && (int)cf->h && (int)sw && (int)sh) {
                cairo_pattern_t * pat = id->cairoImage (cairo_surface);
                if (pat) {
                    cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
                    cairo_save (cr);
                    cairo_matrix_t matrix;
                    cairo_matrix_init_identity (&matrix);
                    float scalex = 1.0 * sw / cf->w;
                    float scaley = 1.0 * sh / cf->h;
                    cairo_matrix_scale (&matrix, scalex, scaley);
                    cairo_matrix_translate (&matrix,
                            1.0*sx/scalex - (double)cf->x,
                            1.0*sy/scaley - (double)cf->y);
                    cairo_pattern_set_matrix (pat, &matrix);
                    cairo_rectangle (cr, cf->x, cf->y, cf->w, cf->h);
                    cairo_set_source (cr, pat);
                    cairo_clip (cr);
                    cairo_paint_with_alpha (cr, 1.0 * cf->progress / 100);
                    cairo_restore (cr);
                }
            }
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Wipe * wipe) {
    //debugLog() << "Visit " << wipe->nodeName() << endl;
    if (wipe->target && wipe->target->id == RP::id_node_image) {
        ImageData *id=convertNode<RP::Image>(wipe->target)->cached_img.data.ptr();
        if (id && !id->isEmpty ()) {
            Single x = wipe->x, y = wipe->y;
            Single tx = x, ty = y;
            Single w = wipe->w, h = wipe->h;
            Single sx = wipe->srcx, sy = wipe->srcy, sw = wipe->srcw, sh = wipe->srch;
            if (!(int)sw)
                sw = id->width();
            if (!(int)sh)
                sh = id->height();
            if (wipe->direction == RP::Wipe::dir_right) {
                Single dx = w * 1.0 * wipe->progress / 100;
                tx = x -w + dx;
                w = dx;
            } else if (wipe->direction == RP::Wipe::dir_left) {
                Single dx = w * 1.0 * wipe->progress / 100;
                tx = x + w - dx;
                x = tx;
                w = dx;
            } else if (wipe->direction == RP::Wipe::dir_down) {
                Single dy = h * 1.0 * wipe->progress / 100;
                ty = y - h + dy;
                h = dy;
            } else if (wipe->direction == RP::Wipe::dir_up) {
                Single dy = h * 1.0 * wipe->progress / 100;
                ty = y + h - dy;
                y = ty;
                h = dy;
            }

            if ((int)w && (int)h) {
                cairo_pattern_t * pat = id->cairoImage (cairo_surface);
                if (pat) {
                    cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
                    cairo_matrix_t matrix;
                    cairo_matrix_init_identity (&matrix);
                    float scalex = 1.0 * sw / wipe->w;
                    float scaley = 1.0 * sh / wipe->h;
                    cairo_matrix_scale (&matrix, scalex, scaley);
                    cairo_matrix_translate (&matrix,
                            1.0*sx/scalex - (double)tx,
                            1.0*sy/scaley - (double)ty);
                    cairo_pattern_set_matrix (pat, &matrix);
                    cairo_set_source (cr, pat);
                    cairo_rectangle (cr, x, y, w, h);
                    cairo_fill (cr);
                }
            }
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::ViewChange * vc) {
    //kdDebug() << "Visit " << vc->nodeName() << endl;
    if (vc->unfinished () || vc->progress < 100) {
        cairo_pattern_t * pat = cairo_pop_group (cr); // from imfl
        cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
        cairo_push_group (cr);
        cairo_save (cr);
        cairo_set_source (cr, pat);
        cairo_paint (cr);
        if ((int)vc->w && (int)vc->h && (int)vc->srcw && (int)vc->srch) {
            cairo_matrix_t matrix;
            cairo_matrix_init_identity (&matrix);
            float scalex = 1.0 * vc->srcw / vc->w;
            float scaley = 1.0 * vc->srch / vc->h;
            cairo_matrix_scale (&matrix, scalex, scaley);
            cairo_matrix_translate (&matrix,
                    1.0*vc->srcx/scalex - (double)vc->x,
                    1.0*vc->srcy/scaley - (double)vc->y);
            cairo_pattern_set_matrix (pat, &matrix);
            cairo_set_source (cr, pat);
            cairo_rectangle (cr, vc->x, vc->y, vc->w, vc->h);
            cairo_fill (cr);
        }
        cairo_pattern_destroy (pat);
        cairo_restore (cr);
    }
}

//-----------------------------------------------------------------------------

namespace KMPlayer {

class KMPLAYER_NO_EXPORT MouseVisitor : public Visitor {
    Matrix matrix;
    NodePtr node;
    unsigned int event;
    int x, y;
    bool handled;
    bool bubble_up;
public:
    MouseVisitor (unsigned int evt, int x, int y);
    KDE_NO_CDTOR_EXPORT ~MouseVisitor () {}
    using Visitor::visit;
    void visit (Node * n);
    void visit (SMIL::Layout *);
    void visit (SMIL::Region *);
    void visit (SMIL::TimedMrl * n);
    void visit (SMIL::MediaType * n);
    void visit (SMIL::Anchor *);
    void visit (SMIL::Area *);

};

} // namespace

KDE_NO_CDTOR_EXPORT
MouseVisitor::MouseVisitor (unsigned int evt, int a, int b)
    : event (evt), x (a), y (b), handled (false), bubble_up (false) {
}

KDE_NO_EXPORT void MouseVisitor::visit (Node * n) {
    debugLog () << "Mouse event ignored for " << n->nodeName () << endl;
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::Layout * layout) {
    Matrix m = matrix;
    matrix = Matrix (layout->surface->xoffset, layout->surface->yoffset,
            layout->surface->xscale, layout->surface->yscale);
    matrix.transform (m);

    NodePtr node_save = node;
    node = layout;
    for (NodePtr r = layout->firstChild (); r; r = r->nextSibling ()) {
        if (r->id == SMIL::id_node_region)
            r->accept (this);
        if (!node->active ())
            break;
    }
    node = node_save;

    matrix = m;
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::Region * region) {
    SRect rect = region->surface->bounds;
    Single rx = rect.x (), ry = rect.y(), rw = rect.width(), rh = rect.height();
    matrix.getXYWH (rx, ry, rw, rh);
    handled = false;
    bool inside = x > rx && x < rx+rw && y > ry && y< ry+rh;
    if (!inside && (event == event_pointer_clicked || !region->has_mouse))
        return;
    Matrix m = matrix;
    matrix = Matrix (rect.x(), rect.y(), 1.0, 1.0);
    matrix.transform (m);
    bubble_up = false;

    bool child_handled = false;
    if (inside)
        for (NodePtr r = region->firstChild (); r; r = r->nextSibling ()) {
            r->accept (this);
            child_handled |= handled;
            if (!node->active ())
                break;
        }
    child_handled &= !bubble_up;
    bubble_up = false;

    int saved_event = event;
    if (node->active ()) {
        bool propagate_listeners = !child_handled;
        if (event == event_pointer_moved) {
            propagate_listeners = true; // always pass move events
            if (region->has_mouse && (!inside || child_handled)) {
                region->has_mouse = false;
                event = event_outbounds;
            } else if (inside && !child_handled && !region->has_mouse) {
                region->has_mouse = true;
                event = event_inbounds;
            }
        }// else // event_pointer_clicked
        if (propagate_listeners) {
            NodeRefListPtr nl = region->listeners (
                    event == event_pointer_moved ? mediatype_attached : event);
            if (nl) {
                for (NodeRefItemPtr c = nl->first(); c; c = c->nextSibling ()) {
                    if (c->data)
                        c->data->accept (this);
                    if (!node->active ())
                        break;
                }
            }
        }
    }
    event = saved_event;
    handled = inside;
    matrix = m;
}

static void followLink (SMIL::LinkingBase * link) {
    debugLog () << "link to " << link->href << " clicked" << endl;
    NodePtr n = link;
    if (link->href.hasPrefix ("#")) {
        SMIL::Smil * s = SMIL::Smil::findSmilNode (link);
        if (s)
            s->jump (link->href.mid (1));
        else
            errorLog() << "In document jumps smil not found" << endl;
    } else
        for (NodePtr p = link->parentNode (); p; p = p->parentNode ()) {
            if (n->mrl () && n->mrl ()->opener == p) {
                p->setState (Node::state_deferred);
                p->mrl ()->setParam (StringPool::attr_src, link->href, 0L);
                p->activate ();
                break;
            }
            n = p;
        }
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::Anchor * anchor) {
    if (event == event_pointer_moved)
        ;// set cursor to hand
    else if (event == event_pointer_clicked)
        followLink (anchor);
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::Area * area) {
    NodePtr n = area->parentNode ();
    if (n->id >= SMIL::id_node_first_mediatype &&
            n->id < SMIL::id_node_last_mediatype) {
        SMIL::MediaType * mt = convertNode <SMIL::MediaType> (n);
        Single x1 = mt->cached_rect.x (), x2 = mt->cached_rect.y ();
        Single w = mt->cached_rect.width (), h = mt->cached_rect.height ();
        matrix.getXYWH (x1, x2, w, h);
        if (area->nr_coords > 1) {
            Single left = area->coords[0].size (mt->cached_rect.width ());
            Single top = area->coords[1].size (mt->cached_rect.height ());
            matrix.getXY (left, top);
            if (x < left || x > left + w || y < top || y > top + h)
                return;
            if (area->nr_coords > 3) {
                Single right = area->coords[2].size (mt->cached_rect.width ());
                Single bottom = area->coords[3].size(mt->cached_rect.height ());
                matrix.getXY (right, bottom);
                if (x > right || y > bottom)
                    return;
            }
        }
        if (event == event_pointer_moved)
            ;//cursor.setShape (Qt::PointingHandCursor);
        else {
            NodeRefListPtr nl = area->listeners (event);
            if (nl)
                for (NodeRefItemPtr c = nl->first(); c; c = c->nextSibling ()) {
                    if (c->data)
                        c->data->accept (this);
                    if (!node->active ())
                        return;
                }
            if (event == event_pointer_clicked && !area->href.isEmpty ())
                followLink (area);
        }
    }
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::TimedMrl * timedmrl) {
    timedmrl->runtime ()->processEvent (event);
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::MediaType * mediatype) {
    if (mediatype->sensitivity == SMIL::MediaType::sens_transparent) {
        bubble_up = true;
        return;
    }
    NodeRefListPtr nl = mediatype->listeners (
            event == event_pointer_moved ? mediatype_attached : event);
    if (nl)
        for (NodeRefItemPtr c = nl->first(); c; c = c->nextSibling ()) {
            if (c->data)
                c->data->accept (this);
            if (!node->active ())
                return;
        }
    if (event != event_pointer_moved)
        visit (static_cast <SMIL::TimedMrl *> (mediatype));
    if (event != event_inbounds && event != event_outbounds) {
      SMIL::RegionBase *r=convertNode<SMIL::RegionBase>(mediatype->region_node);
      if (r && r->surface && r->surface->node && r != r->surface->node)
          return r->surface->node->accept (this);
    }
}

//-----------------------------------------------------------------------------

KDE_NO_CDTOR_EXPORT ViewArea::ViewArea ()
 : cairo_surface (0L),
   aspect (0.0),
   widget (0L),
   video (0L),
   surface (new ViewSurface (this)),
   m_repaint_timer (0),
   m_fullscreen (false) {
    if (!image_data_map)
        image_data_map = new ImageDataMap;
}

KDE_NO_CDTOR_EXPORT ViewArea::~ViewArea () {
    if (cairo_surface)
        cairo_surface_destroy (cairo_surface);
    if (image_data_map->size () == 0) {
        delete image_data_map;
        image_data_map = 0;
    }
    if (m_repaint_timer)
        g_source_remove (m_repaint_timer);
}

KDE_NO_EXPORT void ViewArea::fullScreen () {
    m_fullscreen = !m_fullscreen;
    if (cairo_surface) {
        cairo_surface_destroy (cairo_surface);
        cairo_surface = 0L;
    }
}

KDE_NO_EXPORT void ViewArea::mousePressEvent (int x, int y) {
    if (surface->node) {
        MouseVisitor visitor (event_pointer_clicked, x, y);
        surface->node->accept (&visitor);
    }
}

KDE_NO_EXPORT void ViewArea::mouseDoubleClickEvent () {
}

KDE_NO_EXPORT void ViewArea::syncVisual (const SRect & rect) {
    if (!surface->node) {
        //repaint (QRect(rect.x(), rect.y(), rect.width(), rect.height()), false);
        return;
    }
    int ex = rect.x ();
    if (ex > 0)
        ex--;
    int ey = rect.y ();
    if (ey > 0)
        ey--;
    int ew = rect.width () + 3;
    int eh = rect.height () + 3;
    if (!cairo_surface)
        cairo_surface = cairoCreateSurface (widget, width (), height ());
    CairoPaintVisitor visitor (cairo_surface, SRect (ex, ey, ew, eh));
    surface->node->accept (&visitor);
    //XFlush (qt_xdisplay ());
}

KDE_NO_EXPORT void ViewArea::paintEvent (const Rect & rect) {
    if (surface->node)
        scheduleRepaint (rect.x (), rect.y (), rect.width (), rect.height ());
}

KDE_NO_EXPORT void ViewArea::resizeEvent () {
    if (!video)
        return;
    int x = 0, y = 0;
    int w = width ();
    int h = height ();
    if (w <= 0 || h <= 0)
        return;
    scheduleRepaint (0, 0, w, h);
    if (cairo_surface)
        cairo_xlib_surface_set_size (cairo_surface, w, h);
    // now scale the regions and check if video region is already sized
    bool av_geometry_changed = false;
    surface->bounds = SRect (x, y, w, h);
    if (surface->node && w > 0 && h > 0) {
        m_av_geometry = Rect (0, 0, 0, 0);
        surface->node->handleEvent (new SizeEvent (x, y, w, h, fit_meet));
        av_geometry_changed = (m_av_geometry != Rect (0, 0, 0, 0));
        x = m_av_geometry.x ();
        y = m_av_geometry.y ();
        w = m_av_geometry.width();
        h = m_av_geometry.height();
    } else
        m_av_geometry = Rect (x, y, w, h);
    m_av_geometry = Rect (x, y, w, h);
    if (!av_geometry_changed)
        setAudioVideoGeometry (x, y, w, h, 0L);
}

KDE_NO_EXPORT
void ViewArea::setAudioVideoGeometry (int x, int y, int w, int h, unsigned int * bg_color) {
    if (m_av_geometry.width () > 0 && m_av_geometry.height () > 0)
        scheduleRepaint (m_av_geometry.x (), m_av_geometry.y (),
                m_av_geometry.width (), m_av_geometry.height ());
    if (aspect > 0.01) {
        int w1 = w;
        int h1 = h;
        w = int (h * aspect);
        if (w > w1) {
            h = int (w1 / aspect);
            w = w1;
        }
        x += (w1 - w) / 2;
        y += (h1 - h) / 2;
    }
    Color color (bg_color ? (int)*bg_color : 0);
    m_av_geometry = Rect (x, y, w, h);
    debugLog() << "ViewArea::setAudioVideoGeometry " << x << "," << y << " "
        << w << "x" << h << endl;
    gdk_window_move_resize (video, x, y, w, h);
    gdk_window_set_background (video, (GdkColor *)(GColor)color);
    gdk_window_clear (video);
}

KDE_NO_EXPORT
void ViewArea::setAspect (float a) {
    aspect = a;
    /*if (aspect > 0.01) {
        int w, h, x, y;
        gdk_drawable_get_size (video, &w, &h);
        gdk_window_get_position (video, &x, &y);
        scheduleRepaint (x, y, w, h);
        int w1 = int (h * aspect);
        int h1 = h;
        if (w1 > w) {
            h1 = int (w / aspect);
            w1 = w;
        }
        int xoff = (w - w1) / 2;
        int yoff = (h - h1) / 2;
        if (xoff > 0 && yoff > 0)
            gdk_window_move_resize (video, x + xoff, y + yoff, w1, h1);
    }*/
}

KDE_NO_EXPORT SurfacePtr ViewArea::getSurface (NodePtr node) {
    static_cast <ViewSurface *> (surface.ptr ())->clear ();
    surface->node = node;
    //qApp->postEvent (this, new QResizeEvent (size (), QSize (0, 0)));
    if (m_repaint_timer) {
        g_source_remove (m_repaint_timer);
        m_repaint_timer = 0;
    }
    //m_view->viewer()->resetBackgroundColor ();
    if (node)
        return surface;
    return 0L;
}

KDE_NO_EXPORT void ViewArea::scheduleRepaint (Single x, Single y, Single w, Single h) {
    if (m_repaint_timer)
        m_repaint_rect = m_repaint_rect.unite (SRect (x, y, w, h));
    else {
        m_repaint_rect = SRect (x, y, w, h);
        m_repaint_timer = g_timeout_add (20, cb_scheduledPaint, this);
    }
}

KDE_NO_EXPORT
void ViewArea::moveRect (Single x, Single y, Single w, Single h, Single x1, Single y1) {
}

KDE_NO_EXPORT void ViewArea::timeoutPaintEvent () {
    static Single zero;
    m_repaint_timer = 0;
    if (m_repaint_rect.width () > zero && m_repaint_rect.height () > zero) {
        syncVisual (m_repaint_rect);
        if (!m_repaint_timer) {
            m_repaint_timer = g_timeout_add (40, cb_scheduledPaint, this);
            m_repaint_rect = SRect ();
        } else {
            warningLog () << "scheduleRepaint during syncVisual" << endl;
        }
    }
}

KDE_NO_EXPORT int ViewArea::width () const {
    int w, h;
    gdk_drawable_get_size (widget->window, &w, &h);
    return w;
}

KDE_NO_EXPORT int ViewArea::height () const {
    int w, h;
    gdk_drawable_get_size (widget->window, &w, &h);
    return h;
}

GdkFilterReturn videoWidgetFilter (GdkXEvent *xe, GdkEvent *e, gpointer d) {
    XEvent *event = (XEvent *) xe;
        switch (event->xany.type) {
            case KeyPress:
            case KeyRelease:
                if (event->xany.window == GDK_DRAWABLE_XID (GDK_DRAWABLE (
                               ((Application*) d)->viewarea->videoWindow ()))) {
                    XKeyEvent keyevent = event->xkey;
                    keyevent.window =
                        GDK_DRAWABLE_XID (GDK_DRAWABLE (((Application*) d)->viewarea->viewWidget ()->window));
                    XSendEvent (event->xany.display, keyevent.window, true,
                            KeyPressMask, (XEvent *) &keyevent);
                    return GDK_FILTER_REMOVE;
                }
        }
    return GDK_FILTER_CONTINUE;
}

KDE_NO_EXPORT void ViewArea::setOutputWidget (Application *app, GtkWidget * w) {
    widget = w;
    gdk_window_clear (w->window);
    if (cairo_surface) {
        cairo_surface_destroy (cairo_surface);
        cairo_surface = 0L;
    }
    if (!video) {
        GdkWindowAttr attributes;
        attributes.window_type = GDK_WINDOW_CHILD;
        attributes.x = w->allocation.x;
        attributes.y = w->allocation.y;
        attributes.width = w->allocation.width;
        attributes.height = w->allocation.height;
        attributes.wclass = GDK_INPUT_OUTPUT;
        attributes.visual = gtk_widget_get_visual (w);
        attributes.colormap = gtk_widget_get_colormap (w);
        attributes.event_mask = gtk_widget_get_events (w);
        attributes.event_mask &= ~(GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);
        //    ~(GDK_KEY_PRESS_MASK | GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);
        gint mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
        video = gdk_window_new (w->window, &attributes, mask);
        gdk_window_add_filter (video, videoWidgetFilter, app);
        gdk_window_show (video);

        debugLog() << "video window "  << (int)GDK_DRAWABLE_XID (GDK_DRAWABLE (video)) << " " << endl;

    } else
        gdk_window_reparent (video, widget->window, 0, 0);
    gdk_window_focus (widget->window, CurrentTime);
    //resizeEvent ();
}

//------------------------------------------------------------------------------

void cb_view_area_button_press (GtkWidget *w, GdkEventButton *e, Application * app) {
        //debugLog () << "button press " << event->x << "," << event->y << endl;
    app->viewarea->mousePressEvent (int (e->x), int (e->y));
}

void cb_view_area_expose (GtkWidget *w, GdkEventExpose *e, Application * app) {
    //debugLog () << "cb_view_area_expose " << event->area.x << "," << event->area.x << " " << event->area.width << "x" << event->area.height << endl;
    app->viewarea->scheduleRepaint (
            e->area.x, e->area.y, e->area.width, e->area.height);
}

void cb_view_area_realized (GtkWidget *widget, Application * app) {
    if (widget == app->view_area_fullscreen) {
        Color c (0x101010);
        gdk_window_set_background (widget->window, (GdkColor *)(GColor)c);
    } else
        app->viewarea->setOutputWidget (app, widget);
    
}

gboolean cbViewAreaFullscreenMapped (GtkWidget *w, GdkEvent *, Application *a) {
    gtk_window_fullscreen (GTK_WINDOW (a->window_fullscreen));
    a->viewarea->setOutputWidget (a, a->view_area_fullscreen);
    cb_view_area_configure (w, 0L, a);
    return false;
}

void cbViewAreaFullscreenUnmapped (GtkWidget *w, Application *a) {
    a->viewarea->setOutputWidget (a, a->view_area);
    cb_view_area_configure (w, 0L, a);
}

gboolean cb_scheduledPaint (void * vp) {
    static_cast <ViewArea *> (vp)->timeoutPaintEvent ();
    return false;
}


//#ifdef gtk < 2.8

unsigned char * pixbufBits (GdkPixbuf *pixbuf, int & n_channels) {
    gint width = gdk_pixbuf_get_width (pixbuf);
    gint height = gdk_pixbuf_get_height (pixbuf);
    guchar *gdk_pixels = gdk_pixbuf_get_pixels (pixbuf);
    int gdk_rowstride = gdk_pixbuf_get_rowstride (pixbuf);
    n_channels = gdk_pixbuf_get_n_channels (pixbuf);
    guchar *cairo_pixels, *head;
    static cairo_user_data_key_t key;
    int j;

    cairo_pixels = (guchar*)g_malloc (4 * width * height);
    head = cairo_pixels;

    for (j = height; j; j--) {
        guchar *p = gdk_pixels;
        guchar *q = cairo_pixels;

        if (n_channels == 3) {
            guchar *end = p + 3 * width;

            while (p < end) {
#if __BYTE_ORDER == __LITTLE_ENDIAN
                q[0] = p[2];
                q[1] = p[1];
                q[2] = p[0];
                q[3] = p[1];
#warning little ending
#else     
                q[1] = p[0];
                q[2] = p[1];
                q[3] = p[2];
#warning big ending
#endif
                p += 3;
                q += 4;
            }
        } else {
            guchar *end = p + 4 * width;
            guint t1,t2,t3;

#define MULT(d,c,a,t) G_STMT_START { t = c * a + 0x7f; d = ((t >> 8) + t) >> 8; } G_STMT_END

            while (p < end) {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
                MULT(q[0], p[2], p[3], t1);
                MULT(q[1], p[1], p[3], t2);
                MULT(q[2], p[0], p[3], t3);
                q[3] = p[3];
#else     
                q[0] = p[3];
                MULT(q[1], p[0], p[3], t1);
                MULT(q[2], p[1], p[3], t2);
                MULT(q[3], p[2], p[3], t3);
#endif
              
              p += 4;
              q += 4;
            }
          
#undef MULT
        }

        gdk_pixels += gdk_rowstride;
        cairo_pixels += 4 * width;
    }
    return head;
}

/*
void
gdk_cairo_set_source_pixbuf (cairo_t   *cr,
                             GdkPixbuf *pixbuf,
                             double     pixbuf_x,
                             double     pixbuf_y)
{
  gint width = gdk_pixbuf_get_width (pixbuf);
  gint height = gdk_pixbuf_get_height (pixbuf);
  guchar *gdk_pixels = gdk_pixbuf_get_pixels (pixbuf);
  int gdk_rowstride = gdk_pixbuf_get_rowstride (pixbuf);
  int n_channels = gdk_pixbuf_get_n_channels (pixbuf);
  debugLog() << "gdk_cairo_set_source_pixbuf pixbuf:" << width << "x" << height << " rowstride:" << gdk_rowstride << " channels:" << n_channels << endl;
  guchar *cairo_pixels;
  cairo_format_t format;
  cairo_surface_t *surface;
  static cairo_user_data_key_t key;
  int j;

  if (n_channels == 3)
    format = CAIRO_FORMAT_RGB24;
  else
    format = CAIRO_FORMAT_ARGB32;

  cairo_pixels = (guchar*)g_malloc (4 * width * height);
  surface = cairo_image_surface_create_for_data ((unsigned char *)cairo_pixels,
                                                 format,
                                                 width, height, 4 * width);
  cairo_surface_set_user_data (surface, &key,
                               cairo_pixels, (cairo_destroy_func_t)g_free);

  for (j = height; j; j--)
    {
      guchar *p = gdk_pixels;
      guchar *q = cairo_pixels;

      if (n_channels == 3)
        {
          guchar *end = p + 3 * width;
          
          while (p < end)
            {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
              q[0] = p[2];
              q[1] = p[1];
              q[2] = p[0];
#else     
              q[1] = p[0];
              q[2] = p[1];
              q[3] = p[2];
#endif
              p += 3;
              q += 4;
            }
        }
      else
        {
          guchar *end = p + 4 * width;
          guint t1,t2,t3;
            
#define MULT(d,c,a,t) G_STMT_START { t = c * a + 0x7f; d = ((t >> 8) + t) >> 8; } G_STMT_END

          while (p < end)
            {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
              MULT(q[0], p[2], p[3], t1);
              MULT(q[1], p[1], p[3], t2);
              MULT(q[2], p[0], p[3], t3);
              q[3] = p[3];
#else     
              q[0] = p[3];
              MULT(q[1], p[0], p[3], t1);
              MULT(q[2], p[1], p[3], t2);
              MULT(q[3], p[2], p[3], t3);
#endif
              
              p += 4;
              q += 4;
            }
          
#undef MULT
        }

      gdk_pixels += gdk_rowstride;
      cairo_pixels += 4 * width;
    }

  cairo_set_source_surface (cr, surface, pixbuf_x, pixbuf_y);
  cairo_surface_destroy (surface);
}
*/
